/* Copyright (c) 1994,1997 HD Associates (contact: dufault@hda.com)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * This product includes software developed by HD Associates
 * 4. Neither the name of the HD Associaates nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY HD ASSOCIATES``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL HD ASSOCIATES OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* i2c: I^2-C bus support.  This supports a single master talking
 * to multiple slaves.
 *
 * I2C is the Philips two wire serial bus.  The signals are CLOCK
 * and DATA.  See any I2C peripheral spec for an overview.
 *
 * This is a nicer version than the old program.  It should work
 * on fast processors, while the timing of the old one was dependant
 * on the processor speed.  HOWEVER, on slow processors you may want
 * to define the "nanodelay" functions to not do anything since the
 * overhead of the proper protocol may be adequate.
 */

#include "config.h"
#if (N_I2C > 0)

#include "i2c.h"

/* bit banging I2C implementation.  This assumes
 * that both the data and clock signals are in the same register.
 */

/* You need these macros in i2csite.h:
 *
 * I2C_IMAGE:      An image of the output i2c register.
 *                 This is for a register with bits that you must preserve.
 *                 When undefined a local static is used.
 * I2C_IMAGE_LOCK():
 * I2C_IMAGE_UNLOCK():
 *                 Macros that lock/unlock the I2C_IMAGE register.
 *                 These can't be defined if I2C_IMAGE isn't defined.
 *
 * I2C_CLOCK:      Output bit for the i2c clock.
 * I2C_DATA:       Output bit for the i2c data.
 * I2C_OUT(val):   A macro or function that will send out to the i2c port.
 * I2C_IN():       A macro that will get from the i2c port.
 *
 * Optionally:
 *
 * I2C_CLOCK_IN:   input bit for the i2c clock.
 * I2C_DATA_IN:    input bit for the i2c data.
 *
 * When undefined these are defined as I2C_CLOCK and I2C_DATA.
 */

#include "i2csite.h"

/* Default for everything for easy debugging:
 */
#if !defined(I2C_IMAGE)
	static int I2C_IMAGE;
	#define I2C_IMAGE_LOCK()
	#define I2C_IMAGE_UNLOCK()
#endif

#if !defined(I2C_DATA)
	#define I2C_DATA (0x01)
#endif

#if !defined(I2C_CLOCK)
	#define I2C_CLOCK (0x02)
#endif

#if !defined(I2C_OUT)
	#include <stdio.h>
	#define I2C_OUT(VAL) do { \
	printf("\rC: %d D: %d", \
	(VAL & I2C_CLOCK) ? 1 : 0, (VAL & I2C_DATA) ?  1 : 0); \
	fflush(stdout); \
	} while (0)
#endif

#if !defined(I2C_IN)
	#define I2C_IN() I2C_IMAGE
#endif

#if !defined(I2C_DATA_IN)
	#define I2C_DATA_IN I2C_DATA
#endif

#if !defined(I2C_CLOCK_IN)
	#define I2C_CLOCK_IN I2C_CLOCK
#endif

#define I2C_SET(BIT) do { \
	I2C_IMAGE_LOCK(); \
	I2C_IMAGE |= BIT; \
	I2C_IMAGE_UNLOCK(); \
	I2C_OUT(I2C_IMAGE); \
} while (0)

#define I2C_CLR(BIT) do { \
	I2C_IMAGE_LOCK(); \
	I2C_IMAGE &= ~(BIT); \
	I2C_IMAGE_UNLOCK(); \
	I2C_OUT(I2C_IMAGE); \
} while (0)

/* It is the responsibility of the code that detects an error
 * to free up the bus.  None of the macros do that, so I'm putting
 * them in upper case to remind us it is a macro even though they
 * have no side effects.
 */
#define CLOCK_RELEASE() I2C_SET(I2C_CLOCK)
#define CLOCK_DRIVE()    I2C_CLR(I2C_CLOCK)

#define DATA_RELEASE()  I2C_SET(I2C_DATA)
#define DATA_DRIVE()     I2C_CLR(I2C_DATA)

#define DATA_NOT_DRIVEN()    (I2C_IN() & I2C_DATA_IN)
#define DATA_IS_DRIVEN()     (!DATA_NOT_DRIVEN())

#define CLOCK_IS_DRIVEN()    (!(I2C_IN() & I2C_CLOCK_IN))
#define CLOCK_NOT_DRIVEN()    (!CLOCK_IS_DRIVEN())

#define BUS_FREE() \
	((I2C_IN() & (I2C_CLOCK_IN|I2C_DATA_IN)) == (I2C_CLOCK_IN|I2C_DATA_IN))

#if !defined(MAX_SLAVE_DELAY)
	#define MAX_SLAVE_DELAY 10000000
#endif


#include "nano.h"

static u_char in_use;

/* i2c_init: Be sure both clock ad data are released.
 */
void i2c_init(void)
{
	CLOCK_RELEASE();
	DATA_RELEASE();
#if defined(I2C_SLAVE)
	i2cslave_init();
#endif
}

/* give_up_bus; No longer using the bus.  This will
 * return the appropriate error to return to the next level.
 */
static int give_up_bus(const char *s)
{
	in_use = 0;

	CLOCK_RELEASE();
	DATA_RELEASE();

	if (s)
	{
		return 1;
	}

	return 0;
}

/* Clock release: Release the clock, delay the minimum clock high
 * time, and then synchronize with the slowest slave device on the bus
 * by watching for the clock to go high when it stops driving it.
 *
 * In violation of the bus spec (which has NO maximum)
 * we only permit a maximum time (see MAX_SLAVE_DELAY)
 * for the sloweset device.
 */
static int clock_release(void)
{
	nanotimer_t t;

	CLOCK_RELEASE();
	nanodelay(I2C_T_high);
	
	t = nanotimer_set(MAX_SLAVE_DELAY);

	while (!nanotimer_elapsed(&t) && CLOCK_IS_DRIVEN())
		;

	if (CLOCK_IS_DRIVEN())
		return give_up_bus("jammed");

	return 0;
}

/* send_bit: Send a bit.
 *
 * For consistency with the i2c ACK condition, return 0 for success
 * and 1 for failure.  A 1 means we lost the bus arbitration.
 *
 * Start:             Clock is driven low, data is not driven
 * End                Clock is driven low, data is not driven
 */
static int send_bit(u_char value)
{
	/* Center the data bit in the I2C_T_low pulse:
	 */

	nanodelay(I2C_T_low_2);

	if (value == 0)
		DATA_DRIVE();

	nanodelay(I2C_T_low_2);

	if (clock_release())
		return 1;

	/* If we sent a high bit (we're not driving the bus) and someone else
	 * is then we're not really the master and we lost arbitration.
	 */

	if (value)
	{
		if (DATA_IS_DRIVEN())
			return give_up_bus("collision");
	}

	CLOCK_DRIVE();

	nanodelay(I2C_T_f);

	DATA_RELEASE();

	return 0;
}

/* recv_bit: Receive a bit.
 * Start:             Clock is driven low, data is not driven
 * End                Clock is driven low, data is not driven
 */
static int recv_bit(u_char *vp, u_char mask)
{
	nanodelay(I2C_T_low);

	if (clock_release())
		return 1;

	if (DATA_NOT_DRIVEN())
		(*vp) |= mask;
	else
		(*vp) &= ~mask;

	CLOCK_DRIVE();

	return 0;
}

/* i2c_put: Put a value on the I2C bus.  Return the ACK
 * from the slave.  This is 0 for OK and 1 for failure.
 */
int i2c_put(u_char value)
{
	u_char ack = 0x00, mask;

	for (mask = 0x80; mask != 0x00; mask >>= 1)
		if (send_bit(value & mask))
			return 1;

	if (recv_bit(&ack, 0x01))
		ack = 1;

	return (int)ack;
}

/* i2c_start: Send a start condition out the bus.  This is a high to
 *            low data transition when the clock is high.
 *
 * For consistency with the i2c ACK condition, return 0 for success
 * and 1 for failure.
 *
 * Start              Bus is free (clock and data not driven)
 * End                Clock is driven low, data is not driven
 */
int i2c_start(u_char slave)
{
	nanotimer_t t;

	/* Are we already in use?
	 */

	if (in_use)
	{
		return 1;
	}
	in_use = 1;

	/* Poll the bus for the bus free time to verify it is really free:
	 */

	t = nanotimer_set(I2C_T_buf);

	while (BUS_FREE() && !nanotimer_elapsed(&t))
		;

	if (!BUS_FREE())
		return give_up_bus("never_free");

	/* The start condition is a high to low data transition while
  	 *  the clock is high followed by a delay:
	 */

	DATA_DRIVE();
	nanodelay(I2C_T_hd_sta);		/* Start condition hold time */

	/* We may own the bus now.
	 *
	 * NOTE WELL: In a multimaster situation you
	 * must watch the result of each bit you send to
	 * be sure that when you send a high bit it shows up high
	 * on the bus - bus arbitration continues forever by watching
	 * for a collision.  Whichever master device
	 * sends the first non-identical low bit wins arbitration.
	 *
	 * Drive the clock low so all devices know the bus is busy, delay
	 * the signal fall time , and release the data signal.
	 */
	CLOCK_DRIVE();
	nanodelay(I2C_T_f);			/* signal fall time */
	DATA_RELEASE();

	if (i2c_put(slave))
	{
		(void)i2c_stop();
		return 1;
	}

	return 0;
}


/* i2c_stop: Generate a stop condition.
 * Start:             Clock is driven low, data is not driven
 * End                Bus is free (clock and data not driven)
 */
int i2c_stop(void)
{
	nanodelay(I2C_T_low_2);		/* Delay low time */
	DATA_DRIVE();			/* Ensure data is low */
	nanodelay(I2C_T_low_2);		/* Delay low time */

	if (clock_release() == 0)
	{
		DATA_RELEASE();	/* Stop condition */
		give_up_bus(0);
		return 0;
	}

	return 1;
}


/* i2c_get_no_ack: Get a byte but don't send an ACK.
 */
int i2c_get_no_ack(u_char *p)
{
	u_char value = 0x00, mask;

	for (mask = 0x80; mask != 0x00; mask >>= 1)
		if (recv_bit(&value, mask))
			return 1;

	if (p)
		*p = value;

	return 0;
}

/* i2c_get: Get a byte and send an ACK.
 */
int i2c_get(u_char *p)
{
	if (i2c_get_no_ack(p) == 0)
		return send_bit(0);        /* ACK */

	return 1;
}

/* i2c_write: Write to a given I2C slave on the bus.
 */
ssize_t i2c_write(int slave, const void *p, size_t n)
{
	const u_char *buff = (const u_char *)p;
	size_t i;

	if (i2c_start(slave))
		return 0;

	in_use = slave;

	for (i = 0; i < n; i++)
		if (i2c_put(*buff++))
		{
			(void)give_up_bus("put_fail");
			return i;
		}

	i2c_stop();

	return n;
}

/* i2c_read: Read from a given I2C slave on the bus.
 */
ssize_t i2c_read(int slave, void *p, size_t n)
{
	u_char *buff = (u_char *)p;
	size_t i;

	if (i2c_start(slave | 0x01))
		return 0;

	in_use = slave;

	for (i = 0; i < n - 1; i++)
	{
		if (i2c_get(buff++))
		{
			(void)give_up_bus("get_fail");
			return i;
		}
	}

	if (i2c_get_no_ack(buff++))
	{
		(void)give_up_bus("get_fail");
		return i;
	}

	i2c_stop();

	return n;
}

#if defined(I2C_SLAVE)

/* I2C slave implementation.  This must be polled at 2.35 microseconds
 * for standard I2C.
 */

enum bit_state
{
	/* ack, idle, stop: These are all idle conditions.  They represent
	 * the condition that terminated the last transfer.
	 */
	Bit_idle,
	Bit_after_ack,
	Bit_after_stop,

	/* wait_clock_high_data, clock_low_data: Wait for the end of a
	 * clock cycle assocated with a bit transfer.
	 */
	Bit_wait_clock_high_data,
	Bit_wait_clock_low_data,

	/* wait_clock_high_ack, clock_low_ack: Wait for the end of a clock
	 * cycle associated with an ACK transfer
	 */
	Bit_wait_clock_high_ack,
	Bit_wait_clock_low_ack,

	Bit_unknown
};

struct bit
{
	enum bit_state state;	/* The state of the bit state machine
							 */

	int bit;				/* The bit we're assembling */
	u_char ibuf;			/* Input data buffer */
	u_char obuf;			/* Output data buffer */

	bool data_was_driven;	/* The data was driven on the last poll */
	bool ack;				/* ACK-NAK for the last tranfer */
};

struct bus
{
	enum i2cslave_bus_state state;
	u_char slave;
	u_char transferred;

	/* When transmitting or acknowledging I keep asserting the
	 * data each time through the state machine.  This is needed
	 * if you want to be able to "talk to yourself" in loopback
	 * mode.
	 */
	int slave_transmitter;
	int acknowledge;

	struct bit bit;
};

#if 1
#define i2cslave_debug(ARGS) do {} while(0)
#else
#define i2cslave_debug(ARGS) fprintf ARGS
#endif

/* bit_state_machine: handle the state transitions associated with
 * receiving and sending bits.
 */
static void bit_state_machine(struct bus *bus)
{
	bool data_is_driven = DATA_IS_DRIVEN();
	bool clock_is_driven = CLOCK_IS_DRIVEN();

	switch(bus->bit.state)
	{
		case Bit_after_ack:
		case Bit_after_stop:
		case Bit_idle:
		break;

		case Bit_wait_clock_high_data:

		if (clock_is_driven)
		{
			/* Are we supposed to transmit this bit?
			 */
			if (bus->slave_transmitter)
			{
				if ((bus->bit.obuf & bus->bit.bit) == 0)
					DATA_DRIVE();
				else
					DATA_RELEASE();
			}
		}
		else
		{
			/* Latch the data when the clock goes high:
			 */
			if (data_is_driven)
				bus->bit.ibuf &= ~bus->bit.bit;
			else
				bus->bit.ibuf |= bus->bit.bit;

			bus->bit.bit >>= 1;

			bus->bit.state = Bit_wait_clock_low_data;
		}
		break;

		case Bit_wait_clock_low_data:

		/* If the clock is still high and we saw a low-high transition
		 * on the data we have a stop condition.
		 */

		if (!clock_is_driven && bus->bit.data_was_driven && !data_is_driven)
		{
			i2cslave_debug((stderr, "STOP at bit %02x\n", bus->bit.bit));

			bus->bit.state = Bit_after_stop;
			break;
		}
		else if (clock_is_driven)
		{
			/* If we're transmitting release the data bus.
			 */
			if (bus->slave_transmitter)
				DATA_RELEASE();

			if (bus->bit.bit == 0)	/* Have we transmitted all data bits? */
				bus->bit.state = Bit_wait_clock_high_ack;
			else
				bus->bit.state = Bit_wait_clock_high_data;
		}
		break;

		case Bit_wait_clock_high_ack:
		if (!clock_is_driven)
		{
			bus->bit.ack = !data_is_driven;
			i2cslave_debug((stderr, "ACK %d\n", bus->bit.ack));
			bus->bit.state = Bit_wait_clock_low_ack;
		}
		else if (bus->acknowledge)
			DATA_DRIVE();

		break;

		case Bit_wait_clock_low_ack:
		/* If the clock is still high and we saw a low-high transition
		 * on the data we have a stop condition.
		 */

		if (!clock_is_driven && bus->bit.data_was_driven && !data_is_driven)
		{
			i2cslave_debug((stderr, "STOP at ACK\n"));
			bus->bit.state = Bit_after_stop;
			bus->slave_transmitter = 0;
			bus->acknowledge = 0;
			break;
		}
		else if (clock_is_driven)
		{
			bus->bit.state = Bit_after_ack;

			if (bus->acknowledge)
			{
				DATA_RELEASE();
				bus->acknowledge = 0;
			}
			bus->slave_transmitter = 0;
		}

		break;

		default:
		i2cslave_debug((stderr, "Unknown state.\n"));
		abort();
	}

	bus->bit.data_was_driven = data_is_driven;
}

static struct bus bus_store, *bus = &bus_store;

void i2cslave_init(void)
{
	bus->state = I2cbus_free;
	bus->bit.state = Bit_idle;
}

/* i2cslave_acknowledge: Accept the connection as a slave.  This must
 * be called immediately after getting the "I2cbus_get_slave_addr" event.
 */
void i2cslave_acknowledge()
{
	bus->acknowledge = 1;
}

/* i2cslave_send: Send data as a slave-transmitter.  This must be called after
 * getting the "I2cbus_transfer" event.
 */
void i2cslave_send(int data)
{
	bus->bit.obuf = data;
	bus->slave_transmitter = 1;
	bus->acknowledge = 1;	/* The slave-transmitter still generates the ACK */
}

enum i2cslave_bus_state i2cslave_poll(int *arg)
{
	enum i2cslave_bus_state transition = I2cbus_none;

	bit_state_machine(bus);

	switch(bus->state)
	{
		case I2cbus_free:
		if (CLOCK_NOT_DRIVEN() && DATA_IS_DRIVEN())
		{
			i2cslave_debug((stderr, "Start.\n"));
			transition = bus->state = I2cbus_start;
		}
		break;

		case I2cbus_start:
		if (CLOCK_IS_DRIVEN() && DATA_IS_DRIVEN())	/* He owns the bus */
		{
			transition = bus->state = I2cbus_get_slave_addr;

			bus->bit.ibuf = 0;
			bus->bit.bit = 0x80;

			bus->bit.state = Bit_wait_clock_high_data;
		}
		break;

		case I2cbus_get_slave_addr:
		{
			switch(bus->bit.state)
			{
				case Bit_wait_clock_high_ack:
				bus->slave = bus->bit.ibuf;
				transition = bus->state = I2cbus_got_slave_addr;
				i2cslave_debug((stderr, "Addressed %02x.\n", bus->slave));

				if (arg)
					*(int *)arg = bus->slave;

				break;

				default:
			}
		}
		break;

		case I2cbus_got_slave_addr:
		{
			switch(bus->bit.state)
			{
				case Bit_after_stop:
				transition = bus->state = I2cbus_free;
				break;

				case Bit_after_ack:
				if (bus->slave & 0x01)	/* Reading */
					i2cslave_debug((stderr, "Reading from %02x.\n", bus->slave));
				else
					i2cslave_debug((stderr, "Writing to %02x.\n", bus->slave));

				transition = bus->state = I2cbus_transfer;
				bus->bit.state = Bit_idle;

				break;

				default:
			}
		}
		break;

		case I2cbus_transfer:
			switch(bus->bit.state)
			{
				case Bit_idle:

				bus->bit.ibuf = 0;
				bus->bit.bit = 0x80;

				bus->bit.state = Bit_wait_clock_high_data;
				break;

				case Bit_wait_clock_high_ack:
				bus->transferred = bus->bit.ibuf;

				bus->bit.ibuf = 0;
				bus->bit.bit = 0x80;

				i2cslave_debug(
				(stderr, "Transferred %02x.\n", bus->transferred));

				if (arg)
					*(int *)arg = bus->transferred;

				transition = bus->state = I2cbus_transferred;
				break;

				case Bit_after_stop:
				transition = bus->state = I2cbus_free;
				break;

				default:
			}
			break;

		case I2cbus_transferred:
			switch(bus->bit.state)
			{
				case Bit_after_ack:
				bus->bit.bit = 0x80;
				bus->bit.state = Bit_wait_clock_high_data;
				transition = bus->state = I2cbus_transfer;
				break;

				case Bit_after_stop:
				transition = bus->state = I2cbus_free;
				break;

				default:
			}
			break;

		default:
			i2cslave_debug((stderr, "Unknown case.\n"));
			abort();
	}

	return transition;
}

#endif /* I2C_SLAVE */

#endif	/* N_I2C */
