#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include "i2c.h"
#include "nano.h"

/* Note: You DO NOT need any of this "i2cslave" stuff unless you
 * wish to act as a slave on the bus.  I'm using this to drive
 * the simulated bus through its paces as a demonstration and test.
 * A real master-only needs to only provide the hooks defined in
 * i2csite.h and nano.h.  Stub versions for i2csite.h are in i2c.c.
 */

static struct
{
	u_char slave;
	u_char cmd[32];
	int at;
} transact;

static void
i2c_transaction(void)
{

	fprintf(stdout, "\r\ndevice %d: \"%s\"\n", transact.slave, transact.cmd);

	transact.at = 0;
}

/* slave_poll: Poll the i2c bus to drive the slave mode.
 */
void slave_poll(int sig)
{
	int arg;
	enum i2cslave_bus_state state;
	static int receiving, sending;

	/* Mask off simulated interrupts so we don't go nesting down
	 * into the simulator:
	 */

	switch(state = i2cslave_poll(&arg))
	{
		/* Here is how you would send back data:
		 */
#define MYADDR 0x44
		case I2cbus_got_slave_addr:
		if (arg & 0x01)
		{
			switch(arg)
			{
				case (0x01 | MYADDR):
				i2cslave_acknowledge();
				sending = 1;
				break;

				default:
				fprintf(stdout,
				"\r\nNot generating data for device %02x.\n", arg);
			}
		}
		else
		{
			fprintf(stdout, "\r\nReceiving data for device %02x.\n", arg);
			receiving = 1;
			transact.slave = arg;
			i2cslave_acknowledge();
		}
		break;

		case I2cbus_transfer:
		if (sending)
			i2cslave_send(0xab);
		break;

		case I2cbus_transferred:
		if (receiving)
		{
			if (transact.at < sizeof(transact.cmd) / sizeof(transact.cmd[0]))
				transact.cmd[transact.at++] = arg;

			i2cslave_acknowledge();
		}
		break;

		case I2cbus_free:
		if (receiving)
		{
			i2c_transaction();
			receiving = 0;
		}
		if (sending)
			sending = 0;

		default:
		break;
	}
}


/* Although usleep is supposed to restart any pending alarms
 * it isn't.  Just buzz.
 */

static void buzz(u_long t)
{
	volatile u_long b = t * 100;

	while (b-- != 0)
		;
}

int main(int argc, char *argv[])
{
	ssize_t n;
	char s[] = "Hello world";

	i2c_init();
	i2cslave_init();

	ualarm(100, 100);
	signal(SIGALRM, slave_poll);

	if( (n = i2c_write(8, s, sizeof(s))) != sizeof(s))
		fprintf(stderr, "wrote %d not %d.\n", n, sizeof(s));

	buzz(1000);	/* Make sure we let the polling state machine see this.  */

	return errno;
}

/* Here are the three nanotimer functions you must provide:
 */
nanotimer_t nanotimer_set(u_long t)
{
	return t;
}

int nanotimer_elapsed(nanotimer_t *t)
{
	if (1)
		buzz(1);
	return (*t)-- == 0;
}

void nanodelay(u_long t)
{
	if (1)
		buzz(1 * t);
}
